package webit.script.util;
// Copyright (c) 2003-2013, Jodd Team (jodd.org). All Rights Reserved.
//package jodd.io;

/**
 * Fast, fast
 * <code>char</code> buffer with some additional features.
 */
public class FastCharBuffer implements CharSequence, Appendable {

    // @@generated
    private char[][] buffers = new char[16][];
    private int buffersCount;
    private int currentBufferIndex = -1;
    private char[] currentBuffer;
    private int offset;
    private int count;

    /**
     * Creates a new
     * <code>char</code> buffer. The buffer capacity is initially 1024 bytes,
     * though its size increases if necessary.
     */
    public FastCharBuffer() {
        this(1024);
    }

    /**
     * Creates a new
     * <code>char</code> buffer, with a buffer capacity of the specified size,
     * in bytes.
     *
     * @param size the initial size.
     * @throws IllegalArgumentException if size is negative.
     */
    public FastCharBuffer(int size) {
        if (size < 0) {
            throw new IllegalArgumentException(StringUtil.concat("Invalid size: ", size));
        }
        needNewBuffer(size);
    }

    private void needNewBuffer(int newCount) {
        if (currentBufferIndex < buffersCount - 1) {	// recycling old buffer
            offset = 0;
            currentBufferIndex++;
            currentBuffer = buffers[currentBufferIndex];
        } else {										// creating new buffer
            int newBufferSize;
            if (currentBuffer == null) {
                newBufferSize = newCount;
            } else {
                newBufferSize = Math.max(
                        currentBuffer.length << 1,
                        newCount - count);		// this will give no free additional space

            }

            currentBufferIndex++;
            currentBuffer = new char[newBufferSize];
            offset = 0;

            // add buffer
            if (currentBufferIndex >= buffers.length) {
                int newLen = buffers.length << 1;
                char[][] newBuffers = new char[newLen][];
                System.arraycopy(buffers, 0, newBuffers, 0, buffers.length);
                buffers = newBuffers;
            }
            buffers[currentBufferIndex] = currentBuffer;
            buffersCount++;
        }
    }

    /**
     * Appends
     * <code>char</code> array to buffer.
     */
    public FastCharBuffer append(char[] array, int off, int len) {
        int end = off + len;
        if ((off < 0)
                || (off > array.length)
                || (len < 0)
                || (end > array.length)
                || (end < 0)) {
            throw new IndexOutOfBoundsException();
        }
        if (len == 0) {
            return this;
        }
        int newCount = count + len;
        int remaining = len;
        while (remaining > 0) {
            int part = Math.min(remaining, currentBuffer.length - offset);
            System.arraycopy(array, end - remaining, currentBuffer, offset, part);
            remaining -= part;
            offset += part;
            count += part;
            if (remaining > 0) {
                needNewBuffer(newCount);
            }
        }
        return this;
    }

    /**
     * Appends
     * <code>char</code> array to buffer.
     */
    public FastCharBuffer append(char[] array) {
        return append(array, 0, array.length);
    }

    /**
     * Appends single
     * <code>char</code> to buffer.
     */
    public FastCharBuffer append(char element) {
        if (offset == currentBuffer.length) {
            needNewBuffer(count + 1);
        }

        currentBuffer[offset] = element;
        offset++;
        count++;

        return this;
    }

    /**
     * Appends another fast buffer to this one.
     */
    public FastCharBuffer append(FastCharBuffer buff) {
        for (int i = 0; i < buff.currentBufferIndex; i++) {
            append(buff.buffers[i]);
        }
        append(buff.currentBuffer, 0, buff.offset);
        return this;
    }

    /**
     * Returns buffer size.
     */
    public int size() {
        return count;
    }

    /**
     * Tests if this buffer has no elements.
     */
    public boolean isEmpty() {
        return count == 0;
    }

    /**
     * Returns current index of inner
     * <code>char</code> array chunk. Represents the index of last used inner
     * array chunk.
     */
    public int index() {
        return currentBufferIndex;
    }

    /**
     * Returns the offset of last used element in current inner array chunk.
     */
    public int offset() {
        return offset;
    }

    /**
     * Returns
     * <code>char</code> inner array chunk at given index. May be used for
     * iterating inner chunks in fast manner.
     */
    public char[] array(int index) {
        return buffers[index];
    }

    /**
     * Resets the buffer content.
     */
    public void clear() {
        count = 0;
        offset = 0;
        currentBufferIndex = 0;
        currentBuffer = buffers[currentBufferIndex];
        buffersCount = 1;
    }

    /**
     * Creates
     * <code>char</code> array from buffered content.
     */
    public char[] toArray() {
        int remaining = count;
        int pos = 0;
        char[] array = new char[count];
        for (char[] buf : buffers) {
            int c = Math.min(buf.length, remaining);
            System.arraycopy(buf, 0, array, pos, c);
            pos += c;
            remaining -= c;
            if (remaining == 0) {
                break;
            }
        }
        return array;
    }

    /**
     * Creates
     * <code>char</code> array from buffered content.
     */
    public char[] toArraySkipIfLeftNewLine() {
        if (count > 0) {
            final int skip;
            char[] first = buffers[0];
            if (first[0] == '\n') {
                skip = 1;
            } else if (first[0] == '\r') {
                if (count > 1) {
                    if (first[1] == '\n') {
                        skip = 2;
                    } else {
                        skip = 1;
                    }
                } else {
                    return new char[0];
                }
            } else {
                skip = 0;
            }
            if (skip == 0) {
                return toArray();
            } else {
                //
                int remaining = count - skip;
                int pos = 0;
                char[] array = new char[remaining];
                //first
                int c = Math.min(first.length - skip, remaining);
                System.arraycopy(first, skip, array, pos, c);
                pos += c;
                remaining -= c;
                for (int i = 1; remaining > 0 && i < buffers.length;) {
                    char[] buf = buffers[i++];
                    c = Math.min(buf.length, remaining);
                    System.arraycopy(buf, 0, array, pos, c);
                    pos += c;
                    remaining -= c;
                }
                return array;
            }
        } else {
            return new char[0];
        }
    }

    /**
     * Creates
     * <code>char</code> subarray from buffered content.
     */
    public char[] toArray(int start, int len) {
        int remaining = len;
        int pos = 0;
        char[] array = new char[len];

        if (len == 0) {
            return array;
        }

        int i = 0;
        while (start >= buffers[i].length) {
            start -= buffers[i].length;
            i++;
        }

        while (i < buffersCount) {
            char[] buf = buffers[i];
            int c = Math.min(buf.length - start, remaining);
            System.arraycopy(buf, start, array, pos, c);
            pos += c;
            remaining -= c;
            if (remaining == 0) {
                break;
            }
            start = 0;
            i++;
        }
        return array;
    }

    /**
     * Returns
     * <code>char</code> element at given index.
     */
    public char get(int index) {
        if (index >= count) {
            throw new IndexOutOfBoundsException();
        }
        int ndx = 0;
        while (true) {
            char[] b = buffers[ndx];
            if (index < b.length) {
                return b[index];
            }
            ndx++;
            index -= b.length;
        }
    }

    // @@generated
    /**
     * Returns buffer length, same as {@link #size()}.
     */
    public int length() {
        return count;
    }

    /**
     * Returns buffer content as a string.
     */
    public String toString() {
        return new String(toArray());
    }

    /**
     * Returns char at given index.
     */
    public char charAt(int index) {
        return get(index);
    }

    /**
     * Returns sub sequence.
     */
    public CharSequence subSequence(int start, int end) {
        int len = end - start;
        return new StringBuilder(len).append(toArray(start, len));
    }

    // ---------------------------------------------------------------- additional appenders
    /**
     * Appends string content to buffer.
     */
    public FastCharBuffer append(String string) {
        return append(string.toCharArray());
    }

    /**
     * Appends character sequence to buffer.
     */
    public FastCharBuffer append(CharSequence csq) {
        append(csq, 0, csq.length());
        return this;
    }

    /**
     * Appends character sequence to buffer.
     */
    public FastCharBuffer append(CharSequence csq, int start, int end) {
        for (int i = start; i < end; i++) {
            append(csq.charAt(i));
        }
        return this;
    }

    public void trimRightBlankToNewLine() {
        int tmp_offset;
        int tmp_count = this.count;
        char[] tmp_buf; // = this.currentBuffer;
        int tmp_currentBufferIndex = this.currentBufferIndex;
        boolean notLastOne = false;
        for (; tmp_currentBufferIndex >= 0; tmp_currentBufferIndex--) {
            tmp_buf = buffers[tmp_currentBufferIndex];
            if (notLastOne) {
                tmp_offset = tmp_buf.length;
            } else {
                tmp_offset = this.offset;
                notLastOne = true;
            }
            int pos = CharUtil.lastNotWhitespaceOrNewLine(tmp_buf, 0, tmp_offset);
            if (pos < 0) {
                //All blank
                tmp_count -= tmp_offset;
            } else if (tmp_buf[pos] == '\n' || tmp_buf[pos] == '\r') {
                count = tmp_count - tmp_offset + pos + 1;
                offset = pos + 1;
                currentBufferIndex = tmp_currentBufferIndex;
                currentBuffer = tmp_buf;
                return;
            } else {
                //Not new Line
                break;
            }
        }
    }
}
